/*
   Usage:
     QGPArmbot arm;
     arm.begin(1, 10, 9, 6);
     arm.openGripper();
     arm.gotoPoint(-80, 100, 140);
     arm.closeGripper();
     arm.gotoPoint(70, 200, 10);
     arm.openGripper();
*/
#include <Arduino.h>
#include "ik.h"
#include "QGPArmbot.h"
#include <Servo.h>

int QGP_armMode = 0;

bool setup_servo(ServoInfo &svo, const int n_min, const int n_max,
                 const float a_min, const float a_max)
{
  float n_range = n_max - n_min;
  float a_range = a_max - a_min;

  // Must have a non-zero angle range
  if (a_range == 0)
    return false;

  // Calculate gain and zero
  svo.gain = n_range / a_range;
  svo.zero = n_min - svo.gain * a_min;

  // Set limits
  svo.n_min = n_min;
  svo.n_max = n_max;

  return true;
}

int angle2pwm(const ServoInfo &svo, const float angle)
{
  float pwm = 0.5f + svo.zero + svo.gain * angle;
  return int(pwm);
}

//Full constructor with calibration data
QGPArmbot::QGPArmbot(int sweepMinBase, int sweepMaxBase, float angleMinBase, float angleMaxBase,
                     int sweepMinShoulder, int sweepMaxShoulder, float angleMinShoulder, float angleMaxShoulder,
                     int sweepMinElbow, int sweepMaxElbow, float angleMinElbow, float angleMaxElbow,
                     int sweepMinGripper, int sweepMaxGripper, float angleMinGripper, float angleMaxGripper)
{
  //calroutine();
  setup_servo(_svoBase, sweepMinBase, sweepMaxBase, angleMinBase, angleMaxBase);
  setup_servo(_svoShoulder, sweepMinShoulder, sweepMaxShoulder, angleMinShoulder, angleMaxShoulder);
  setup_servo(_svoElbow, sweepMinElbow, sweepMaxElbow, angleMinElbow, angleMaxElbow);
  setup_servo(_svoGripper, sweepMinGripper, sweepMaxGripper, angleMinGripper, angleMaxGripper);
}

void QGPArmbot::begin(int pinBase, int pinElbow, int pinShoulder, int pinGripper)
{
  _pinBase = pinBase;
  _pinShoulder = pinShoulder;
  _pinElbow = pinElbow;
  _pinGripper = pinGripper;
  _base.attach(_pinBase);
  _shoulder.attach(_pinShoulder);
  _elbow.attach(_pinElbow);
  _gripper.attach(_pinGripper);

  goDirectlyTo(0, 150, 80);
  //goDirectlyToCylinder(0, 100, 50);
  openGripper();
}

void QGPArmbot::end()
{
  _base.detach();
  _shoulder.detach();
  _elbow.detach();
  _gripper.detach();
}

//Set servos to reach a certain point directly without caring how we get there
void QGPArmbot::goDirectlyTo(float x, float y, float z)
{
  float radBase, radShoulder, radElbow;
  if (solve(x, y, z, radBase, radShoulder, radElbow))
  {
    // Serial.print(radBase);
    // Serial.print(",");
    _base.write(angle2pwm(_svoBase, radBase));
    // Serial.print(radShoulder);
    // Serial.print(",");
    _shoulder.write(angle2pwm(_svoShoulder, radShoulder));
    // Serial.print(radElbow);
    // Serial.println(" ");
    _elbow.write(angle2pwm(_svoElbow, radElbow));
    _x = x;
    _y = y;
    _z = z;
  }
}

//Travel smoothly from current point to another point
void QGPArmbot::gotoPoint(float x, float y, float z)
{
  //Starting points - current pos
  float x0 = _x;
  float y0 = _y;
  float z0 = _z;
  float dist = sqrt((x0 - x) * (x0 - x) + (y0 - y) * (y0 - y) + (z0 - z) * (z0 - z));
  int step = 10;
  for (int i = 0; i < dist; i += step)
  {
    goDirectlyTo(x0 + (x - x0) * i / dist, y0 + (y - y0) * i / dist, z0 + (z - z0) * i / dist);
    delay(50);
  }
  goDirectlyTo(x, y, z);
  delay(50);
}

//Get x and y from theta and r
void QGPArmbot::polarToCartesian(float theta, float r, float &x, float &y)
{
  _r = r;
  _t = theta;
  x = r * sin(theta);
  y = r * cos(theta);
}

//Same as above but for cylindrical polar coodrinates
void QGPArmbot::gotoPointCylinder(float theta, float r, float z)
{
  float x, y;
  polarToCartesian(theta, r, x, y);
  gotoPoint(x, y, z);
}

void QGPArmbot::goDirectlyToCylinder(float theta, float r, float z)
{
  float x, y;
  polarToCartesian(theta, r, x, y);
  goDirectlyTo(x, y, z);
}

//Check to see if possible
bool QGPArmbot::isReachable(float x, float y, float z)
{
  float radBase, radShoulder, radElbow;
  return (solve(x, y, z, radBase, radShoulder, radElbow));
}

//Grab something
void QGPArmbot::openGripper(int gp)
{
  _gripper.write(gp); //angle2pwm(_svoGripper,0)
  _grp = gp;
  // delay(300);
}

//Let go of something
void QGPArmbot::closeGripper(int gp)
{
  _gripper.write(gp); //angle2pwm(_svoGripper,90)
  _grp = gp;
  // delay(300);
}

void QGPArmbot::record()
{
  if ((_eex == _x && _eey == _y && _eez == _z && _eegrp == _grp) || _grp == 999)
  {
    return;
  }
  PT pt = {
      getX(), getY(), getZ(), _grp};
  EEPROM.put(eeAddress+4, pt);
  _eex = _x;
  _eey = _y;
  _eez = _z;
  _eegrp = _grp;
  eeAddress += PT_LEN;
  EEPROM.write(0,eeAddress);
}

void QGPArmbot::replay(int _loop, int _delay)
{
  // Serial.println("===REPLAY====");
  byte _eeAddress=EEPROM.read(0);
  do
  {
    //     Serial.print("loop:");
    // Serial.println(_loop);
    cli();
    sei();
    for (int i = 4; i < _eeAddress; i += PT_LEN)
    {
      if (QGP_armMode != ARM_MODE_REPLAY)
      {
        return;
      }
      PT pt; //Variable to store custom object read from EEPROM.
      EEPROM.get(i, pt);
      gotoPoint(pt.x, pt.y, pt.z);
      _gripper.write(pt.grp);
      // Serial.print(i);
      // Serial.print(":");
      // Serial.print(pt.x);
      // Serial.print(",");
      // Serial.print(pt.y);
      // Serial.print(",");
      // Serial.print(pt.z);
      // Serial.print("\n");
      delay(10);
    }
    
    delay(_delay);
    if (_loop >0){
      _loop--;
    }else if(_loop==0){
      break;
    }
  } while (true);
}

void QGPArmbot::moveXYZG(int xyzg, int n)
{
  if (xyzg == AXIS_X)
  {
    goDirectlyTo(getX() + n, getY(), getZ());
  }
  else if (xyzg == AXIS_Y)
  {
    goDirectlyTo(getX(), getY() + n, getZ());
  }
  else if (xyzg == AXIS_Z)
  {
    goDirectlyTo(getX(), getY(), getZ() + n);
  }
  else if (xyzg == AXIS_G)
  {
    if (_grp >= 0 && _grp <= 180)
    {
      _grp += n;
      _gripper.write(_grp);
    }
  }
  delay(10);
}

//Current x, y and z
float QGPArmbot::getX()
{
  return _x;
}
float QGPArmbot::getY()
{
  return _y;
}
float QGPArmbot::getZ()
{
  return _z;
}

float QGPArmbot::getR()
{
  return _r;
}
float QGPArmbot::getTheta()
{
  return _t;
}

int QGPArmbot::getArmMode()
{
  return QGP_armMode;
}

void QGPArmbot::setArmMode(int _mode)
{
  QGP_armMode = _mode;
}

void QGPArmbot::cleanRecord()
{
  for (uint16_t i = 0; i < EEPROM.length(); i++)
  {
    EEPROM.write(i, 0);
  }
  eeAddress = 0;
}